/*
 * This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ********
 * Simple log module
 *
 * stderr is used as output
 * 5 log levels : debug, info, warning, critical and error
 * 5 macros defined (one per log level), usage à la printf
 *   e.g.: xlog_debug("created %d objects", object_cout);
 *         xlog_info("Download speed %4.1lf %s", transfert_speed, transfert_units);
 *         xlog_warning("Hash table 80% full !");
 *         xlog_critical("Core dump expected soon");
 *         xlog_error("Unable to allocate memory\n->%s", streeror(errno));
 * Debug, info, warning and critical macros are no-op if XLOG is not defined
 * Error macro always aborts
 *
 * log record format :
 *
 * Rec ID|L|domain|file|line|function|cpu|elapsed|PID|threadid|message
 *
 * Rec ID   : unsigned int starting at 0, incremented each time a record is
 *            created. Incrementation is atomic (no need to mutex the acces).
 *            If the program is forked, each copy of the process has its own
 *            record id incremented independently.
 * L        : char (log level)
 *            one of D (debug), I (info), W (warning), C (critical), E (error)
 *            Debug level message are emitted only if DEBUG macro is defined
 *            Error level message causes program abort
 * domain   : value of the __DOMAIN__ macro at compilation time
 * file     : file name where the call occured
 * line     : position of the call
 * function : caller function
 * cpu      : cpu time used in seconds
 * elapsed  : time elapsed from the start of the program in seconds
 * PID      : PID from the calling process
 * threadid : Thread ID from the calling thread
 *            always 0 if THREAD macro is undefined
 * message  : Free message
 *
 **************************
 * NEVER directly use xlog_log or xlog_level_t in your code as they are not defined
 * if XLOG is not.
 */


#ifndef xlog_h__
#define xlog_h__

#ifdef XLOG

#  include <stdarg.h>
#  include "xgnu_extensions.h"

typedef enum { XLOG_LEVEL_DEBUG,
               XLOG_LEVEL_INFO, XLOG_LEVEL_WARNING, XLOG_LEVEL_CRITICAL,
               XLOG_LEVEL_ERROR
             } xlog_level_t;
#define LEVEL_CHAR "DIWCE"

void xlog_log (const char*, const char*, int, const char*, xlog_level_t, const char*, ...) XGNU_PRINTF(6,7) XGNU_NONNULL(1,2,4,6);

#  ifndef __DOMAIN__
#    define __DOMAIN__ ""
#  endif

#  ifndef __FILE_BASENAME__
#    define __FILE_BASENAME__ __FILE__
#  endif

#  ifdef DEBUG
#    define xlog_debug(format, ...)   xlog_log( __DOMAIN__, __FILE_BASENAME__, __LINE__, __func__, XLOG_LEVEL_DEBUG,   format, ##__VA_ARGS__ )
#    define xlog_debug_if(expr, format, ...) ({if (expr) {xlog_debug(format, ##__VA_ARGS__);}})
#  else /* !defined(DEBUG) */
#    define xlog_debug(format, ...)
#    define xlog_debug_if(expr, format, ...)
#  endif /* defined(DEBUG) */

#  define xlog_info(format, ...)     xlog_log( __DOMAIN__, __FILE_BASENAME__, __LINE__, __func__, XLOG_LEVEL_INFO,     format, ##__VA_ARGS__ )
#  define xlog_warning(format, ...)  xlog_log( __DOMAIN__, __FILE_BASENAME__, __LINE__, __func__, XLOG_LEVEL_WARNING,  format, ##__VA_ARGS__ )
#  define xlog_critical(format, ...) xlog_log( __DOMAIN__, __FILE_BASENAME__, __LINE__, __func__, XLOG_LEVEL_CRITICAL, format, ##__VA_ARGS__ )
#  define xlog_error(format, ...)    xlog_log( __DOMAIN__, __FILE_BASENAME__, __LINE__, __func__, XLOG_LEVEL_ERROR,    format, ##__VA_ARGS__ )

#  define xlog_info_if(expr, format, ...)     ({if (expr) {xlog_info(format, ##__VA_ARGS__);}})
#  define xlog_warning_if(expr, format, ...)  ({if (expr) {xlog_warning(format, ##__VA_ARGS__);}})
#  define xlog_critical_if(expr, format, ...) ({if (expr) {xlog_critical(format, ##__VA_ARGS__);}})
#  define xlog_error_if(expr, format, ...)    ({if (expr) {xlog_error(format, ##__VA_ARGS__);}})

#else /* !defined(XLOG) */

#  include <stdio.h>
#  include <stdlib.h>
#  include "xgnu_extensions.h"
#  include "xbacktrace.h"

#  define xlog_debug(format, ...)
#  define xlog_info(format, ...)
#  define xlog_warning(format, ...)
#  define xlog_critical(format, ...)
#  define xlog_error(format, ...)   \
          do {                                                                 \
                fprintf(stderr,"%s(%s):%d -> ", __FILE__, __func__, __LINE__); \
                fprintf(stderr,format, ##__VA_ARGS__);                         \
                fprintf(stderr,"\naborting\n");                                \
                xbacktrace();                                                  \
    				    exit(EXIT_FAILURE);                                            \
                XGNU_UNREACHABLE;                                              \
              } while(0)

#  define xlog_debug_if(format, ...)
#  define xlog_info_if(expr, format, ...)
#  define xlog_warning_if(expr, format, ...)
#  define xlog_critical_if(expr, format, ...)
#  define xlog_error_if(expr, format, ...) ({if(expr){xlog_error(format,##__VA_ARGS__);}})

#endif /* XLOG */

#endif /* xlog_h__ */
